#
#  __version__.py
#
#  Copyright (c) 2016-2023 Junpei Kawamoto
#
#  This file is part of rgmining-fraudar.
#
#  rgmining-fraudar is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  rgmining-fraudar is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with rgmining-fraudar. If not, see <http://www.gnu.org/licenses/>.
#
from typing import Final

__version__: Final = "0.7.0"



# FRAUDAR: Bounding Graph Fraud in the Face of Camouflage
# Authors: Bryan Hooi, Hyun Ah Song, Alex Beutel, Neil Shah, Kijung Shin, Christos Faloutsos
#
# This software is licensed under Apache License, Version 2.0 (the  "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Version: 1.1
# Date: June 12, 2018
# Main Contact: Bryan Hooi (bhooi@andrew.cmu.edu)

from .MinTree import MinTree


def main():
    T = MinTree([1, 4, 2, 5, 3])
    T.dump()
    T.popMin()
    T.dump()
    T.popMin()
    T.dump()
    T.changeVal(4, 3)
    T.dump()
    T.popMin()
    T.dump()
    T.popMin()
    T.dump()
    T.changeVal(4, 10)
    T.dump()
    T.popMin()
    T.dump()

    T = SamplingTree([10, 40, 12, 50, 30], 1)
    T.dump()
    T.sample()
    T.dump()
    T.sample()
    T.dump()
    T.changeVal(3, 3)
    T.dump()
    T.sample()
    T.dump()
    T.sample()
    T.dump()
    T.changeVal(4, 10)
    T.dump()
    T.sample()
    T.dump()


if __name__ == "__main__":
    main()



#
#  greedy.py
#
#  Copyright (c) 2016-2023 Junpei Kawamoto
#
#  This file is part of rgmining-fraudar.
#
#  rgmining-fraudar is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  rgmining-fraudar is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with rgmining-fraudar. If not, see <http://www.gnu.org/licenses/>.
#
#  This file was originally made by Bryan Hooi et al,
#  and distributed under Apache License, Version 2.0.
#
"""
This module contains functions that run the greedy detector for dense regions in a sparse matrix.
use aveDegree or sqrtWeightedAveDegree or logWeightedAveDegree on a sparse matrix,
which returns ((rowSet, colSet), score) for the most suspicious block.
"""
import random

import numpy as np
from scipy import sparse

from .MinTree import MinTree

np.set_printoptions(linewidth=160)


# given 2 lists corresponding to the edge source and destination,
# this returns the sparse matrix representation of the data
# @profile
def listToSparseMatrix(edgesSource, edgesDest):
    m = max(edgesSource) + 1
    n = max(edgesDest) + 1
    M = sparse.coo_matrix(([1] * len(edgesSource), (edgesSource, edgesDest)), shape=(m, n))
    M1 = M > 0
    return M1.astype("int")


# reads matrix from file and returns sparse matrix. first 2 columns should be row and column indices of ones.
# @profile
def readData(filename):
    # dat = np.genfromtxt(filename, delimiter='\t', dtype=int)
    edgesSource = []
    edgesDest = []
    with open(filename) as f:
        for line in f:
            toks = line.split()
            edgesSource.append(int(toks[0]))
            edgesDest.append(int(toks[1]))
    return listToSparseMatrix(edgesSource, edgesDest)


def detectMultiple(M, detectFunc, numToDetect):
    Mcur = M.copy().tolil()
    res = []
    for i in range(numToDetect):
        ((rowSet, colSet), score) = detectFunc(Mcur)
        res.append(((rowSet, colSet), score))
        (rs, cs) = Mcur.nonzero()
        for i in range(len(rs)):
            if rs[i] in rowSet and cs[i] in colSet:
                Mcur[rs[i], cs[i]] = 0
    return res


# inject a clique of size m0 by n0, with density pp. the last parameter testIdx determines the camouflage type.
# testIdx = 1: random camouflage, with camouflage density set so each fraudster outputs approximately equal number
#              of fraudulent and camouflage edges
# testIdx = 2: random camouflage, with double the density as in the previous setting
# testIdx = 3: biased camouflage, more likely to add camouflage to high degree columns
def injectCliqueCamo(M, m0, n0, p, testIdx):
    (m, n) = M.shape
    M2 = M.copy().tolil()

    colSum = np.squeeze(M2.sum(axis=0).A)
    colSumPart = colSum[n0:n]
    colSumPartPro = np.int_(colSumPart)
    colIdx = np.arange(n0, n, 1)
    population = np.repeat(colIdx, colSumPartPro, axis=0)

    for i in range(m0):
        # inject clique
        for j in range(n0):
            if random.random() < p:
                M2[i, j] = 1
        # inject camo
        if testIdx == 1:
            thres = p * n0 / (n - n0)
            for j in range(n0, n):
                if random.random() < thres:
                    M2[i, j] = 1
        if testIdx == 2:
            thres = 2 * p * n0 / (n - n0)
            for j in range(n0, n):
                if random.random() < thres:
                    M2[i, j] = 1
        # biased camo
        if testIdx == 3:
            colRplmt = random.sample(population, int(n0 * p))
            M2[i, colRplmt] = 1

    return M2.tocsc()


# sum of weighted edges in rowSet and colSet, plus node suspiciousness values, in matrix M
def c2Score(M, rowSet, colSet, nodeSusp):
    suspTotal = nodeSusp[0][list(rowSet)].sum() + nodeSusp[1][list(colSet)].sum()
    return M[list(rowSet), :][:, list(colSet)].sum(axis=None) + suspTotal


def jaccard(pred, actual):
    intersectSize = len(set.intersection(pred[0], actual[0])) + len(set.intersection(pred[1], actual[1]))
    unionSize = len(set.union(pred[0], actual[0])) + len(set.union(pred[1], actual[1]))
    return intersectSize / unionSize


def getPrecision(pred, actual):
    intersectSize = len(set.intersection(pred[0], actual[0])) + len(set.intersection(pred[1], actual[1]))
    return intersectSize / (len(pred[0]) + len(pred[1]))


def getRecall(pred, actual):
    intersectSize = len(set.intersection(pred[0], actual[0])) + len(set.intersection(pred[1], actual[1]))
    return intersectSize / (len(actual[0]) + len(actual[1]))


def getFMeasure(pred, actual):
    prec = getPrecision(pred, actual)
    rec = getRecall(pred, actual)
    return 0 if (prec + rec == 0) else (2 * prec * rec / (prec + rec))


def getRowPrecision(pred, actual, idx):
    intersectSize = len(set.intersection(pred[idx], actual[idx]))
    return intersectSize / len(pred[idx])


def getRowRecall(pred, actual, idx):
    intersectSize = len(set.intersection(pred[idx], actual[idx]))
    return intersectSize / len(actual[idx])


def getRowFMeasure(pred, actual, idx):
    prec = getRowPrecision(pred, actual, idx)
    rec = getRowRecall(pred, actual, idx)
    return 0 if (prec + rec == 0) else (2 * prec * rec / (prec + rec))


# run greedy algorithm using square root column weights
def sqrtWeightedAveDegree(M, nodeSusp=None):
    (m, n) = M.shape
    colSums = M.sum(axis=0)
    colWeights = 1.0 / np.sqrt(np.squeeze(colSums) + 5)
    colDiag = sparse.lil_matrix((n, n))
    colDiag.setdiag(colWeights)
    W = M * colDiag
    return fastGreedyDecreasing(W, colWeights, nodeSusp)


# run greedy algorithm using logarithmic weights
def logWeightedAveDegree(M, nodeSusp=None):
    (m, n) = M.shape
    colSums = M.sum(axis=0)
    colWeights = np.squeeze(np.array(1.0 / np.log(np.squeeze(colSums) + 5)))
    colDiag = sparse.lil_matrix((n, n))
    colDiag.setdiag(colWeights)
    W = M * colDiag
    print("finished computing weight matrix")
    return fastGreedyDecreasing(W, colWeights, nodeSusp)


def aveDegree(M, nodeSusp=None):
    (m, n) = M.shape
    return fastGreedyDecreasing(M, [1] * n, nodeSusp)


def subsetAboveDegree(M, col_thres, row_thres):
    M = M.tocsc()
    (m, n) = M.shape
    colSums = np.squeeze(np.array(M.sum(axis=0)))
    rowSums = np.squeeze(np.array(M.sum(axis=1)))
    colValid = colSums > col_thres
    rowValid = rowSums > row_thres
    M1 = M[:, colValid].tocsr()
    M2 = M1[rowValid, :]
    rowFilter = [i for i in range(m) if rowValid[i]]
    colFilter = [i for i in range(n) if colValid[i]]
    return M2, rowFilter, colFilter


# @profile
def fastGreedyDecreasing(M, colWeights, nodeSusp=None):
    (m, n) = M.shape
    if nodeSusp is None:
        nodeSusp = (np.zeros(m), np.zeros(n))
    Md = M.todok()
    Ml = M.tolil()
    Mlt = M.transpose().tolil()
    rowSet = set(range(0, m))
    colSet = set(range(0, n))
    curScore = c2Score(M, rowSet, colSet, nodeSusp)

    bestAveScore = curScore / (len(rowSet) + len(colSet))
    bestSets = (rowSet, colSet)
    print("finished initialization")
    rowDeltas = (
        np.squeeze(M.sum(axis=1).A) + nodeSusp[0]
    )  # contribution of this row to total weight, i.e. *decrease* in total weight when *removing* this row
    colDeltas = np.squeeze(M.sum(axis=0).A) + nodeSusp[1]
    print("finished setting deltas")
    rowTree = MinTree(rowDeltas)
    colTree = MinTree(colDeltas)
    print("finished building min trees")

    numDeleted = 0
    deleted = []
    bestNumDeleted = 0

    while rowSet and colSet:
        if (len(colSet) + len(rowSet)) % 100000 == 0:
            print("current set size = %d" % (len(colSet) + len(rowSet),))
        (nextRow, rowDelt) = rowTree.getMin()
        (nextCol, colDelt) = colTree.getMin()
        if rowDelt <= colDelt:
            curScore -= rowDelt
            for j in Ml.rows[nextRow]:
                delt = colWeights[j]
                colTree.changeVal(j, -colWeights[j])
            rowSet -= {nextRow}
            rowTree.changeVal(nextRow, float("inf"))
            deleted.append((0, nextRow))
        else:
            curScore -= colDelt
            for i in Mlt.rows[nextCol]:
                delt = colWeights[nextCol]
                rowTree.changeVal(i, -colWeights[nextCol])
            colSet -= {nextCol}
            colTree.changeVal(nextCol, float("inf"))
            deleted.append((1, nextCol))

        numDeleted += 1
        curAveScore = curScore / (len(colSet) + len(rowSet))

        if curAveScore > bestAveScore:
            bestAveScore = curAveScore
            bestNumDeleted = numDeleted

    # reconstruct the best row and column sets
    finalRowSet = set(range(m))
    finalColSet = set(range(n))
    for i in range(bestNumDeleted):
        if deleted[i][0] == 0:
            finalRowSet.remove(deleted[i][1])
        else:
            finalColSet.remove(deleted[i][1])
    return (finalRowSet, finalColSet), bestAveScore



#
#  __init__.py
#
#  Copyright (c) 2016-2023 Junpei Kawamoto
#
#  This file is part of rgmining-fraudar.
#
#  rgmining-fraudar is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  rgmining-fraudar is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with rgmining-fraudar. If not, see <http://www.gnu.org/licenses/>.
#
"""A wrapper of Fraudar algorithm for the review graph mining project.

The Fraudar has been introduced by Bryan Hooi, *et al.* in
ACM SIGKDD 2016 Conference on Knowledge Discovery & Data Mining (KDD 2016).

This package exports :class:`ReviewGraph <graph.ReviewGraph>` class,
which implements interfaces expected in other APIs of
`Review Graph Mining project <https://rgmining.github.io/>`_,
and three sub algorithms used in FRAUDER:

* :meth:`aveDegree <export.greedy.aveDegree>` computes average degree on a matrix,
* :meth:`sqrtWeightedAveDegree <export.greedy.sqrtWeightedAveDegree>`
  computes square-weighted average degree on a matrix,
* :meth:`logWeightedAveDegree <export.greedy.logWeightedAveDegree>`
  computes logarithm-weighted average degree on a matrix.

:meth:`ReviewGraph <graph.ReviewGraph>` takes keyword argument ``algo`` to
be set the sub algorithm to be used.
"""
from typing import Final

from fraudar.__version__ import __version__
from fraudar.export.greedy import aveDegree, logWeightedAveDegree, sqrtWeightedAveDegree
from fraudar.graph import ReviewGraph

__all__: Final = ("ReviewGraph", "aveDegree", "sqrtWeightedAveDegree", "logWeightedAveDegree", "__version__")



#
#  MinTree.py
#
#  Copyright (c) 2016-2023 Junpei Kawamoto
#
#  This file is part of rgmining-fraudar.
#
#  rgmining-fraudar is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  rgmining-fraudar is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with rgmining-fraudar. If not, see <http://www.gnu.org/licenses/>.
#
#  This file was originally made by Bryan Hooi et al,
#  and distributed under Apache License, Version 2.0.
#
import math


class MinTree:
    """
    A tree data structure which stores a list of degrees and can quickly retrieve the min degree element,
    or modify any of the degrees, each in logarithmic time. It works by creating a binary tree with the
    given elements in the leaves, where each internal node stores the min of its two children.
    """

    def __init__(self, degrees):
        self.height = int(math.ceil(math.log(len(degrees), 2)))
        self.numLeaves = 2**self.height
        self.numBranches = self.numLeaves - 1
        self.n = self.numBranches + self.numLeaves
        self.nodes = [float("inf")] * self.n
        for i in range(len(degrees)):
            self.nodes[self.numBranches + i] = degrees[i]
        for i in reversed(range(self.numBranches)):
            self.nodes[i] = min(self.nodes[2 * i + 1], self.nodes[2 * i + 2])

    # @profile
    def getMin(self):
        cur = 0
        for i in range(self.height):
            cur = (2 * cur + 1) if self.nodes[2 * cur + 1] <= self.nodes[2 * cur + 2] else (2 * cur + 2)
        # print "found min at %d: %d" % (cur, self.nodes[cur])
        return (cur - self.numBranches, self.nodes[cur])

    # @profile
    def changeVal(self, idx, delta):
        cur = self.numBranches + idx
        self.nodes[cur] += delta
        for i in range(self.height):
            cur = (cur - 1) // 2
            nextParent = min(self.nodes[2 * cur + 1], self.nodes[2 * cur + 2])
            if self.nodes[cur] == nextParent:
                break
            self.nodes[cur] = nextParent

    def dump(self):
        print(f"numLeaves: {self.numLeaves}, numBranches: {self.numBranches}, n: {self.n}, nodes: ")
        cur = 0
        for i in range(self.height + 1):
            for j in range(2**i):
                print(self.nodes[cur], end=" ")
                cur += 1
            print("")



#
#  __init__.py
#
#  Copyright (c) 2016-2023 Junpei Kawamoto
#
#  This file is part of rgmining-fraudar.
#
#  rgmining-fraudar is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  rgmining-fraudar is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with rgmining-fraudar. If not, see <http://www.gnu.org/licenses/>.
#
#  This file was originally made by Bryan Hooi et al,
#  and distributed under Apache License, Version 2.0 at
#  https://github.com/bhooi/bhooi.github.io/tree/master/code/camo.
#
"""
FRAUDAR: Bounding Graph Fraud in the Face of Camouflage
Authors: Bryan Hooi, Hyun Ah Song, Alex Beutel, Neil Shah, Kijung Shin, Christos Faloutsos

This software is licensed under Apache License, Version 2.0 (the  "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Version: 1.1
Date: June 12, 2018
Main Contact: Bryan Hooi (bhooi@andrew.cmu.edu)
"""

__version__ = "1.1"



# FRAUDAR: Bounding Graph Fraud in the Face of Camouflage
# Authors: Bryan Hooi, Hyun Ah Song, Alex Beutel, Neil Shah, Kijung Shin, Christos Faloutsos
#
# This software is licensed under Apache License, Version 2.0 (the  "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Version: 1.1
# Date: June 12, 2018
# Main Contact: Bryan Hooi (bhooi@andrew.cmu.edu)

import sys

# Runs the greedy algorithm.
# Command line arguments: first argument is path to data file. Second argument is path to save output.
# Third argument (*optional*) is path where the node suspiciousness values are stored.
import time

from .greedy import *


def main():
    start_time = time.time()
    M = readData(sys.argv[1])
    # The example data is a 500 x 500 matrix with an injected dense block among the first 20 nodes
    print("finished reading data: shape = %d, %d @ %d" % (M.shape[0], M.shape[1], time.time() - start_time))

    if len(sys.argv) > 3:  # node suspiciousness present
        print("using node suspiciousness")
        rowSusp = np.loadtxt("%s.rows" % (sys.argv[3],))
        colSusp = np.loadtxt("%s.cols" % (sys.argv[3],))
        lwRes = logWeightedAveDegree(M, (rowSusp, colSusp))
    else:
        lwRes = logWeightedAveDegree(M)

    print(lwRes)
    np.savetxt("%s.rows" % (sys.argv[2],), np.array(list(lwRes[0][0])), fmt="%d")
    np.savetxt("%s.cols" % (sys.argv[2],), np.array(list(lwRes[0][1])), fmt="%d")
    print("score obtained is %f" % (lwRes[1],))
    print("done @ %f" % (time.time() - start_time,))

    # When no node suspiciousness values are passed, we detect the injected block.
    # However, when adding large suspiciousness values to the first 10 rows and columns,
    # the new detected block becomes the first 10 rows and columns.
    # Intuitively, the first 10 rows and columns are now suspiciousness enough that it no longer becomes worth it
    # for the algorithm to include the remaining 10 rows and columns.


if __name__ == "__main__":
    main()



#
#  graph.py
#
#  Copyright (c) 2016-2023 Junpei Kawamoto
#
#  This file is part of rgmining-fraudar.
#
#  rgmining-fraudar is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  rgmining-fraudar is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with rgmining-fraudar. If not, see <http://www.gnu.org/licenses/>.
#
"""Provide a review graph which runs Fraudar algorithm.
"""
import tempfile
from bisect import bisect_left
from collections import defaultdict
from typing import Any, Final, Protocol

import numpy as np

from fraudar.export import greedy
from fraudar.export.greedy import logWeightedAveDegree


class Node:
    """Node of the ReviewGraph.

    A node has a name and a link to the graph. It also implements
    :meth:`__hash__` function so that each node can be stored in
    dictionaries.

    Args:
      graph: graph object this node belongs to.
      name: name of this node.
    """

    graph: Final["ReviewGraph"]
    """The graph object this node belongs to."""
    name: Final[str]
    """Name of this node."""

    __slots__ = ("graph", "name")

    def __init__(self, graph: "ReviewGraph", name: str) -> None:
        """Construct a node instance.

        Args:
          graph: graph object this node belongs to.
          name: name of this node.
        """
        self.graph = graph
        self.name = name

    def __hash__(self) -> int:
        """Returns a hash value of this instance."""
        return 13 * hash(type(self)) + 17 * hash(self.name)

    def __lt__(self, other: "Node") -> bool:
        return self.name.__lt__(other.name)


class Reviewer(Node):
    """A node type representing a reviewer.

    Use :meth:`ReviewGraph.new_reviewer` to create a new reviewer object
    instead of using this constructor directory.

    Args:
      graph: graph object this reviewer belongs to.
      name: name of this reviewer.
    """

    anomalous_score: float
    """anomalous score of this reviewer."""

    __slots__ = ("anomalous_score",)

    def __init__(self, graph: "ReviewGraph", name: str, anomalous_score: float = 0) -> None:
        super().__init__(graph, name)
        self.anomalous_score = anomalous_score


class Product(Node):
    """A node type representing a product.

    Use :meth:`ReviewGraph.new_product` to create a new product object
    instead of using this constructor directory.

    Args:
      graph: graph object this product belongs to.
      name: name of this product.
    """

    __slots__ = ()

    @property
    def summary(self) -> float:
        """Summary of ratings given to this product."""
        reviewers = self.graph.reviews[self].keys()
        ratings = [self.graph.reviews[self][r] for r in reviewers]
        weights = [1 - r.anomalous_score for r in reviewers]
        if sum(weights) == 0:
            return float(np.mean(ratings))
        else:
            return float(np.average(ratings, weights=weights))


class _Writable(Protocol):
    def write(self, s: str, /) -> int:
        ...


class ReviewGraph:
    """ReviewGraph is a simple bipartite graph representing review relation.

    Args:
      blocks: how many blocks to be detected. (default: 1)
      algo: algorithm used in fraudar, chosen from
        :meth:`aveDegree <fraudar.export.greedy.aveDegree>`,
        :meth:`sqrtWeightedAveDegree <fraudar.export.greedy.sqrtWeightedAveDegree>`,
        and
        :meth:`logWeightedAveDegree <fraudar.export.greedy.logWeightedAveDegree>`.
        (default: logWeightedAveDegree)
    """

    reviewers: Final[list[Reviewer]]
    """Collection of reviewers."""
    products: Final[list[Product]]
    """Collection of products."""
    reviews: Final[defaultdict[Product, dict[Reviewer, float]]]
    """Collection of reviews.

    reviews is a dictionary of which key is a product and value is another
    dictionary of which key is a reviewer and value is a rating from the
    reviewer to the product.
    """
    _algo: Final[Any]
    _blocks: Final[int]

    def __init__(self, blocks: int = 1, algo: Any = logWeightedAveDegree) -> None:
        self.reviewers = []
        self.products = []
        self.reviews = defaultdict(dict)

        self._algo = algo
        self._blocks = blocks

    def new_reviewer(self, name: str, **_kwargs: Any) -> Reviewer:
        """Create a new reviewer.

        Args:
          name: name of the new reviewer.

        Returns:
          a new reviewer object.
        """
        r = Reviewer(self, name)
        self.reviewers.append(r)
        return r

    def new_product(self, name: str) -> Product:
        """Create a new product.

        Args:
          name: name of the new product.

        Returns:
          a new product object.
        """
        p = Product(self, name)
        self.products.append(p)
        return p

    def add_review(self, reviewer: Reviewer, product: Product, rating: float, **_kwargs: Any) -> float:
        """Add a review from a reviewer to a product.

        Args:
          reviewer: reviewer who posts the review.
          product: product which receives the review.
          rating: the review score.

        Returns:
          added review score.
        """
        self.reviews[product][reviewer] = rating
        return rating

    def update(self) -> float:
        """Update anomalous scores by running a greedy algorithm.

        Returns:
          0
        """
        with tempfile.NamedTemporaryFile(mode="w") as fp:
            # Store this graph to a temporal file.
            self._store_matrix(fp)
            fp.seek(0)

            # Run greedy algorithm.
            M = greedy.readData(fp.name)
            res = greedy.detectMultiple(M, self._algo, self._blocks)

            # Update anomalous scores.
            for block in res:
                for i in block[0][0]:
                    self.reviewers[i].anomalous_score = 1

        return 0

    def _store_matrix(self, fp: _Writable) -> None:
        """Store this graph as a sparse matrix format.

        Args:
          fp: file-like object where the matrix to be written.
        """
        self.reviewers.sort()
        self.products.sort()
        for p in self.reviews:
            j = bisect_left(self.products, p)
            for r in self.reviews[p]:
                i = bisect_left(self.reviewers, r)
                fp.write("{0} {1}\n".format(i, j))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fraudar', 'fraudar.export']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.24.1,<2.0.0', 'scikit-learn>=1.2.0,<2.0.0']

setup_kwargs = {
    'name': 'rgmining-fraudar',
    'version': '0.7.0',
    'description': 'A wrapper of Fraudar algorithm for Review graph mining project',
    'long_description': 'A wrapper of FRAUDAR algorithm\n==============================\n\n|GPLv3| |Build Status| |Maintainability| |Test Coverage| |Release|\n\n|Logo|\n\nThis package implements a wrapper of\n`FRAUDAR <https://bhooi.github.io/projects/fraudar/index.html>`__\nalgorithm to provide APIs defined in `Review Graph Mining\nproject <https://rgmining.github.io/>`__.\n\nInstallation\n------------\n\nUse ``pip`` to install this package.\n\n.. code:: shell\n\n    $pip install --upgrade rgmining-fraudar\n\nLicense\n-------\n\nThis software is released under The GNU General Public License Version\n3, see\n`COPYING <https://github.com/rgmining/fraudar/blob/master/COPYING>`__\nfor more detail.\n\nThe original FRAUDAR source code is made by `Bryan\nHooi <https://bhooi.github.io/index.html>`__, `Hyun Ah\nSong <http://www.cs.cmu.edu/~hyunahs/>`__, `Alex\nBeutel <http://alexbeutel.com/>`__, `Neil\nShah <http://nshah.net/>`__, `Kijung\nShin <https://kijungs.github.io/>`__, and `Christos\nFaloutsos <http://www.cs.cmu.edu/~christos/>`__, and licensed under\n`Apache License, Version 2.0 <LICENSE-2.0>`__. It is available at\nhttps://bhooi.github.io/projects/fraudar/index.html.\n\n.. |GPLv3| image:: https://img.shields.io/badge/license-GPLv3-blue.svg\n   :target: https://www.gnu.org/copyleft/gpl.html\n.. |Build Status| image:: https://github.com/rgmining/fraudar/actions/workflows/python-lib.yaml/badge.svg\n   :target: https://github.com/rgmining/fraudar/actions/workflows/python-lib.yaml\n.. |Maintainability| image:: https://api.codeclimate.com/v1/badges/4c4c3df79b33f65b77cd/maintainability\n   :target: https://codeclimate.com/github/rgmining/fraudar/maintainability\n.. |Test Coverage| image:: https://api.codeclimate.com/v1/badges/4c4c3df79b33f65b77cd/test_coverage\n   :target: https://codeclimate.com/github/rgmining/fraudar/test_coverage\n.. |Release| image:: https://img.shields.io/badge/release-0.7.0-brightgreen.svg\n   :target: https://pypi.org/project/rgmining-fraudar/\n.. |Logo| image:: https://rgmining.github.io/fraudar/_static/image.png\n   :target: https://rgmining.github.io/fraudar/\n',
    'author': 'Junpei Kawamoto',
    'author_email': 'kawamoto.junpei@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://rgmining.github.io/fraudar/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
